#version 400 core

#define MAXCOUNT 512

const double er = 1.0e10;

uniform bool centering;
uniform bool julia;
uniform int count;
uniform dvec2 inflection[MAXCOUNT];

uniform dvec2 center;
uniform dvec2 radius;
uniform dvec2 aspect;

in vec2 texcoord;

out vec4 colour;

dvec2 cMul(dvec2 a, dvec2 b)
{
  return dvec2(a.x * b.x - a.y * b.y, a.x * b.y + a.y * b.x);
}

dvec2 cSqr(dvec2 a)
{
  return cMul(a, a);
}

dvec4 cMul(dvec4 a, dvec4 b)
{
  return dvec4(cMul(a.xy, b.xy), cMul(a.xy, b.zw) + cMul(a.zw, b.xy));
}

dvec4 cSqr(dvec4 a)
{
  return cMul(a, a);
}

void main()
{
  dvec2 texcoord_ = center + cMul(radius, aspect * dvec2(texcoord));
  double px = double(length(dvec4(radius, radius) * dvec4(dFdx(vec2(aspect) * texcoord), dFdy(vec2(aspect) * texcoord))));
  dvec4 c = dvec4(texcoord_, px, 0.0);
  if (0 < count)
  {
    dvec4 f = dvec4(inflection[0], 0.0, 0.0);
    dvec4 d = c;
    if (! centering)
    {
      d -= f;
    }
    c = cSqr(d) + f;
  }
  for (int i = 1; i < count && i < MAXCOUNT; ++i)
  {
    dvec4 f = dvec4(inflection[i], 0.0, 0.0);
    dvec4 d = c - f;
    c = cSqr(d) + f;
  }
  int n = 0;
  dvec4 z = julia && count != 0 ? c : dvec4(0.0, 0.0, 0.0, 0.0);
        c = julia && count != 0 ? dvec4(inflection[count - 1], 0.0, 0.0) : c;
  for (n = 0; n < 1024; ++n)
  {
    if (dot(z.xy, z.xy) >= er)
      break;
    z = cSqr(z) + c;
  }
  if (dot(z.xy, z.xy) < er)
  {
    colour = vec4(1.0, 0.0, 0.0, 1.0);
    return;
  }
  float r = float(length(z.xy));
  float dr = float(length(z.zw));
  float de = 2.0 * r * log(r) / dr;
  float g = tanh(clamp(de, 0.0, 4.0));
  if (isnan(de) || isinf(de) || isnan(dr) || isinf(dr) || isnan(r) || isinf(r) || isnan(g) || isinf(g))
    g = 1.0;
  colour = vec4(vec3(g), 1.0);
}
